/* ip.h */
/* Copyright INMOS Limited 1990,1991,1992 */
/* Version 1.19 */

#ifndef __NETINET_IP_H
#define __NETINET_IP_H

/*
 * Copyright (c) 1980,1983,1988 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that this notice is preserved and that due credit is given
 * to the University of California at Berkeley. The name of the University
 * may not be used to endorse or promote products derived from this
 * software without specific prior written permission. This software
 * is provided ``as is'' without express or implied warranty.
 */

/*
 * Definitions for internet protocol version 4.
 * Per RFC 791, September 1981.
 */
#define IPVERSION       4

/*
 * Structure of an internet header, naked of options.
 */
struct ip
{
    unsigned int    ip_hl:4,    /* header length */
                    ip_v:4,     /* version */
                    ip_tos:8,   /* type of service */
                    ip_len:16;  /* total length */
    unsigned int    ip_id:16,   /* identification */
                    ip_off:16;  /* fragment offset field */
#define IP_DF 0x4000            /* dont fragment flag */
#define IP_MF 0x2000            /* more fragments flag */
    unsigned int    ip_ttl:8,   /* time to live */
                    ip_p:8,     /* protocol */
                    ip_sum:16;  /* checksum */
    struct in_addr  ip_src, ip_dst;     /* source and dest address */
};

/*
 * Definitions for options.
 */
#define IPOPT_COPIED(o)         ((o)&0x80)
#define IPOPT_CLASS(o)          ((o)&0x60)
#define IPOPT_NUMBER(o)         ((o)&0x1f)

#define IPOPT_CONTROL           0x00
#define IPOPT_RESERVED1         0x20
#define IPOPT_DEBMEAS           0x40
#define IPOPT_RESERVED2         0x60

#define IPOPT_EOL               0       /* end of option list */
#define IPOPT_NOP               1       /* no operation */

#define IPOPT_RR                7       /* record packet route */
#define IPOPT_TS                68      /* timestamp */
#define IPOPT_SECURITY          130     /* provide s,c,h,tcc */
#define IPOPT_LSRR              131     /* loose source route */
#define IPOPT_SATID             136     /* satnet id (now obsolete) */
#define IPOPT_SSRR              137     /* strict source route */

/*
 * Offsets to fields in options other than EOL and NOP.
 */
#define IPOPT_OPTVAL            0       /* option ID */
#define IPOPT_OLEN              1       /* option length */
#define IPOPT_OFFSET            2       /* offset within option */
#define IPOPT_MINOFF            4       /* min value of above */

/*
 * Time stamp option structure.
 */
struct ip_timestamp
{
    unsigned int    ipt_code:8, /* IPOPT_TS */
                    ipt_len:8,  /* size of structure (variable) */
                    ipt_ptr:8,  /* index of current entry */
                    ipt_oflw:4, /* overflow counter */
                    ipt_flg:4;  /* flags, see below */
    union
    {
        unsigned long   ipt_time[1];
        struct ipt_ta
        {
            struct in_addr  ipt_addr;
            unsigned long   ipt_time;
        }               ipt_ta[1];
    }               ipt_data;
#define ipt_time        ipt_data.ipt_time
#define ipt_ta          ipt_data.ipt_ta
};

#endif
