/*{{{  Header*/
/*
 *     Program:       itconfig
 *     Module:        %M%     %W% %G%
 *
 *
 *     Copyright Inmos Limited 1990.
 *
 *     Purpose:       main command routines
 *                    for remote configuration program for TCP/IP transputer
 *                    LAN interface products.
 *
 *     Compatability: BSD UNIX (host) and Inmos 1.0 Beta UDP/IP (target node)
 *                    VAX/VMS running WIN/TCP
 *                    PC NFS
 *                    PC TCP
 *
 *     History
 *        GJT     01/05/90      a start!
 *        GJT     16/05/90      updates for latest spec
 *        GJT     31/10/90      updates for alpha release
 *        GJT     15/02/91      Changed all occurences of EEPROM to NVRAM
 *        GJT     18/03/91      Changes for new editing and link speed option
 *        GJT     05/04/91      Added default host and domain names.
 *        NMGS    17/01/92      Added support for PC/TCP
 */


/*}}}  */

#include <stdio.h>
#include "itconfig.h" /**/
#include "types.h" /**/
#include <string.h>

/*{{{  external functions*/

/*
 * External function defs
 */

extern unsigned char *hex_colon_addr ();
extern unsigned char *dec_dot_addr ();
extern void tidy_string ();
extern void ip_addr_copy ();
extern void addr_toword ();
extern void load_help ();
extern void display_help ();
extern int read_net_NVRAM ();
extern int read_net_RAM ();
extern int read_net_file ();
extern int read_resource_NVRAM ();
extern int read_resource_RAM ();
extern int read_resource_file ();
extern void router_edit_command_help ();
extern void router_edit_help ();
extern BOOL NVRAM_write ();
extern void show_entry ();
extern char *show_type ();
/*}}}  */

extern char * malloc ();
extern char * getenv ();

/*{{{  global variables*/
/*
 * Global vraiables  (external)
 */

extern unsigned int ether_addr[6];  /* Ethernet address of transputer TCP/IP system */

extern char id;  /* Id for comms with transputer TCP/ip system */

extern unsigned char current_config_net[];
extern unsigned char current_config_resource[];
extern int current_net_len ;
extern int current_resource_len;

extern BOOL edited_config;  /* Set to TRUE if parameters are changed */

extern int skt;        /* socket id for comms with transputer TCP/IP system */

/*
 * Internal
 */

static char saved_net_filename[MAX_NAME_LEN];

static unsigned char display_config_net[NET_DATA_SIZE];
static unsigned char display_config_resource[RESOURCE_DATA_BLOCK];
static int display_net_len = 0;
static int display_resource_len = 0;

/*}}}  */

/*{{{  edit network newer*/
/*{{{  init brouter NEW*/

static void init_brouter (router_ptr)
unsigned char *router_ptr;
/*
 * Initialise router table entries
 */
{
  int i;
  unsigned char *save_ptr = router_ptr;

  for (i = 0; i < MAX_ROUTER_ENTRIES; i++)
  {
    int j;


    *router_ptr = 0;
    router_ptr++;
    for (j = 0; j < IP_ADDR_LEN; j++, router_ptr++)
      *router_ptr = 0;
    *router_ptr = 0;
    router_ptr++;
    *router_ptr = 0;
    router_ptr++;
    for (j = 0; j < IP_ADDR_LEN; j++, router_ptr++)
      *router_ptr = 0;
  }

  *save_ptr = (char) END_OF_ROUTER_TABLE;

  return;
}
/*}}}  */
/*{{{  create_router_list NEW*/

static int create_router_list (router_ptr, start, end)

unsigned char *router_ptr;
struct router_entry **start, **end;

/*
 * Copy router entries from byte table to linked list
 */

{
  int index = 0;
  struct router_entry *current_entry, *save_entry;

  *start = (struct router_entry *) NULL;
  current_entry = (struct router_entry *) NULL;
  save_entry = (struct router_entry *) NULL;
  while ((index < MAX_ROUTER_ENTRIES) &&
	 ((*router_ptr & END_OF_ROUTER_TABLE) != END_OF_ROUTER_TABLE))
  {
    int i;

    current_entry = (struct router_entry *) malloc(sizeof(struct router_entry));
    if (index == 0)
      *start = current_entry;
    else
      save_entry->next = current_entry;
    current_entry->next = (struct router_entry *) NULL;
    current_entry->prev = save_entry;
    current_entry->type = (*router_ptr) & TYPE_MASK;
    router_ptr++;
    for (i = 0; i < IP_ADDR_LEN; i++, router_ptr++)
      current_entry->address[i] = ((unsigned int) *router_ptr) & 0XFF;
    current_entry->TOS = ((unsigned int) *router_ptr) & 0XFF;
    router_ptr++;
    current_entry->TOS += ((unsigned int) (*router_ptr & 0XFF)) << 8;
    router_ptr++;
    for (i = 0; i < IP_ADDR_LEN; i++, router_ptr++)
      current_entry->gateway[i] = ((unsigned int) *router_ptr) & 0XFF;
    save_entry = current_entry;
    index++;
  }

  *end = current_entry;

  return(index);
}
/*}}}  */
/*{{{  copy router list NEW*/

static void copy_router_list (router_ptr, start)
unsigned char *router_ptr;
struct router_entry *start;

/*
 * Copy linked list of router entries to byte table
 */

{
  int entry_count = 0;
  struct router_entry *current_item = start;

  init_brouter(router_ptr);

  if (current_item != (struct router_entry *) NULL)
  {  /* Table has someting in it */
    do
    {
      int j;
      struct router_entry *temp_item = current_item ;

      *router_ptr = (char) (current_item->type & 0XFF);
      router_ptr++;
      for (j = 0; j < IP_ADDR_LEN; j++, router_ptr++)
	*router_ptr = (char) (current_item->address[j] & 0XFF);
      *router_ptr = (char) (current_item->TOS & 0XFF);
      router_ptr++;
      *router_ptr = (char) ((current_item->TOS & 0XFF00) >> 8);
      router_ptr++;
      for (j = 0; j < IP_ADDR_LEN; j++, router_ptr++)
	*router_ptr = (char) (current_item->gateway[j] & 0XFF);
      current_item = current_item->next;
      entry_count++;
      free((char *) temp_item);
    } while (current_item != (struct router_entry *) NULL);
  }

  if ((entry_count == 0) ||
     (((*(router_ptr - ROUTER_ENTRY_SIZE)) & END_OF_ROUTER_TABLE) != END_OF_ROUTER_TABLE))
  {
    int i;

    *router_ptr = (char) (END_OF_ROUTER_TABLE | NOT_IN_USE_TYPE);
    router_ptr++;
    for (i = 1; i < ROUTER_ENTRY_SIZE; i++)
    {
      *router_ptr = (char) 0;
      router_ptr++;
    }

  }

  return;
}
/*}}}  */
/*{{{  get_router_entry*/
BOOL get_router_entry (item)
struct router_entry *item;
{
  BOOL ok;
  int i;
  unsigned int read_type;
  unsigned int read_tos;
  unsigned int addr[IP_ADDR_LEN];

  ok = get_router_type(&read_type, FALSE);
  if (ok == TRUE)
  {
    item->type = read_type;
    if ((read_type & DEFAULT_TYPE) == 0)
    {
      ok = parse_addr(addr, "Address", FALSE);
      if (ok == TRUE)
      {
	for (i = 0; i < IP_ADDR_LEN; i++)
	item->address[i] = addr[i];
      }

    }
    else
    {
      for (i = 0; i < IP_ADDR_LEN; i++)
	item->address[i] = 0;
    }

    if ((ok == TRUE) && ((read_type & NO_TOS_TYPE) == 0))
    {
      ok = get_int(&read_tos, "TOS", FALSE);
      if (ok == TRUE)
	item->TOS = read_tos;
    }
    else
      item->TOS = (char) 0;

    if (ok == TRUE)
    {
      ok = parse_addr(addr, "Gateway address", FALSE);
      if (ok == TRUE)
      {
	for (i = 0; i < IP_ADDR_LEN; i++)
	  item->gateway[i] = addr[i];
      }
    }
  }
  return(ok);
}
/*}}}  */
/*{{{  edit_router_entry*/
BOOL edit_router_entry (item)
struct router_entry *item;
{
  BOOL ok;
  BOOL changed = FALSE;
  int i;
  unsigned int save_type, read_type;
  unsigned int save_tos, read_tos;
  unsigned int addr[IP_ADDR_LEN], save_addr[IP_ADDR_LEN];

  read_type = item->type;
  ok = get_router_type(&read_type, TRUE);
  if (ok == TRUE) /* Has changed */
  {
    item->type = read_type;
    edited_config = TRUE;
    changed = TRUE;
  }

  if ((item->type & DEFAULT_TYPE) == 0)
  {
    for (i = 0; i < IP_ADDR_LEN; i++)
      addr[i] = item->address[i];
    ok = parse_addr(addr, "Address", TRUE);
    if (ok == TRUE)
    {
      for (i = 0; i < IP_ADDR_LEN; i++)
	item->address[i] = addr[i];
      edited_config = TRUE;
      changed = TRUE;
    }

  }
  else if (edited_config == TRUE)
  {
    for (i = 0; i < IP_ADDR_LEN; i++)
      item->address[i] = 0;
  }

  if ((item->type & NO_TOS_TYPE) == 0)
  {
    read_tos = item->TOS;
    ok = get_int(&read_tos, "TOS", TRUE);
    if (ok == TRUE)
    {
      item->TOS = read_tos;
      edited_config = TRUE;
      changed = TRUE;
    }
  }
  else if (edited_config == TRUE)
    item->TOS = 0;

  for (i = 0; i < IP_ADDR_LEN; i++)
    addr[i] = item->gateway[i];
  ok = parse_addr(addr, "Gateway address", TRUE);
  if (ok == TRUE)
  {
    for (i = 0; i < IP_ADDR_LEN; i++)
      item->gateway[i] = addr[i];
    edited_config = TRUE;
    changed = TRUE;
  }
  return(changed);
}
/*}}}  */
/*{{{  edit network mega mix*/
/*
 * Edit network data entries
 */
void edit_network ()
{
  /*{{{  variables*/
  int index = CHANGE_OFFSET;  /* Set to start of IP address */
  unsigned int linkops_port;
  int link_speed;
  unsigned int addr[IP_ADDR_LEN];
  char addr_line[MAX_NAME_LEN + 1];
  char name[MAX_NAME_LEN + 1];
  char domain_name[MAX_DOMAIN_LEN + 1];
  unsigned char dummy;
  int i;
  int len;
  struct router_entry *start_router, *end_router, *current_entry;
  int no_router_entries;
  BOOL router_changed = FALSE;
  BOOL router_finished = FALSE;
  BOOL ok;
  int entry_number = 1;
  /*}}}  */

  /* Set up linked list of router entries */
  no_router_entries = create_router_list(&current_config_net[ROUTER_OFFSET],
					     &start_router, &end_router);
  /*{{{  do fixed data*/
  get_config_addr (&current_config_net[index], "IP address");
  index += IP_ADDR_LEN;
  get_config_addr (&current_config_net[index], "Sub net mask address");
  index += IP_ADDR_LEN;
  get_config_addr (&current_config_net[index], "Broadcast address");
  index += IP_ADDR_LEN;
  get_config_addr (&current_config_net[index], "Name server 1 address");
  index += IP_ADDR_LEN;
  get_config_addr (&current_config_net[index], "Name server 2 address");
  index += IP_ADDR_LEN;
  
  linkops_port = (unsigned int) ((current_config_net[index] << 8) + current_config_net[index + 1]);
  ok = get_int(&linkops_port, "Linkops port", TRUE);
  if (ok == TRUE)
  {
    edited_config = TRUE;
    current_config_net[index] = (unsigned char) ((linkops_port & 0XFF00) >> 8);  /* MS byte */
    current_config_net[index + 1] = (unsigned char) (linkops_port & 0XFF);  /* LS byte */
  }
  index += 2;  /* Skip over port number */
  
  /* Link speed */
  link_speed = ((int) index_to_speed(current_config_net[index])) & 0XFF;
  do
  {
    ok  = get_int((unsigned int *) &link_speed, "Link speed", TRUE);
  } while (speed_to_index(link_speed) < 0);
  
  if (ok == TRUE)
  {
    current_config_net[index] = (unsigned char) (speed_to_index(link_speed) & 0XFF);
    index++;
    edited_config = TRUE;
  }
  else
    index++;
  current_config_net[index++] = '\0';   /* One bytes of padding */
  
  get_config_name(&current_config_net[index], "TCPlink system name", 
		  MAX_NAME_LEN, DEFAULT_HOST_NAME);
  index = DOMAIN_NAME_OFFSET; /* Set to point to domain name */
  
  get_config_name(&current_config_net[index], "Domain name", 
		  MAX_DOMAIN_LEN, DEFAULT_DOMAIN_NAME);
  index += MAX_DOMAIN_LEN + 1;
  
  /* Word align start of router table */
  for (; ((index % 4) != 0); current_config_net[index] = '\0', index++)
    ;
  /*}}}  */

  /* And now the router table */

  /*{{{  do router table*/
  printf("\nRouter table entries\n\n");
  current_entry = start_router;
  entry_number = 1;
  while (! router_finished)
  {
    struct router_entry *temp_entry;
  
  
    /*{{{  get command*/
    if (current_entry != NULL)
    {
      show_entry(entry_number, current_entry);
      printf("Enter command [E(dit), D(elete), G(oto), I(nsert), N(ew),\n");
      printf("               P(revious), Q(uit), [RETURN](next), ?]:    ");
    }
    else
    {
      printf("\n%d: No current entry\n", entry_number);
      printf("Enter command [I(nsert), N(ew), Q(uit), ?]: ");
    }
    fgets(addr_line, sizeof(addr_line), stdin);
    tidy_string(addr_line);
    /*}}}  */
  
    /*{{{  help*/
    if (addr_line[0] == '?')
      router_edit_command_help();
    /*}}}  */
    /*{{{  quit*/
    else if ((addr_line[0] == 'Q') || (addr_line[0] == 'q'))
      router_finished = TRUE;
    /*}}}  */
    /*{{{  return - move to next entry*/
    else if (addr_line[0] == '\0')
    {
      if (current_entry == NULL)
	; /* SKIP */
      else if (current_entry->next != NULL)
      {
	current_entry = current_entry->next;
	entry_number++;
      }
    }
    /*}}}  */
    /*{{{  move to previous entry*/
    else if ((addr_line[0] == 'P') || (addr_line[0] == 'p'))
    {
      if (current_entry == NULL)
	; /* SKIP */
      else if (current_entry->prev != NULL)
      {
	current_entry = current_entry->prev;
	entry_number--;
      }
    }
    /*}}}  */
    /*{{{  delete entry*/
    else if (((addr_line[0] == 'D') || (addr_line[0] == 'd')) &&
	     (current_entry != NULL))
    {
      struct router_entry *temp_entry;
    
      if (no_router_entries == 0)
	printf("Error: no entries to delete\n");  /* Nothing in table, report and SKIP */
      else if (current_entry->prev == (struct router_entry *) NULL)
      { /* start of list */
    
	temp_entry = current_entry;
	current_entry = current_entry->next;
	free((char *) temp_entry);
	if (current_entry != NULL)  /* Not last item in list */
	  current_entry->prev = NULL;
	else
	  end_router = NULL; /* Last item in list, set end pointer to null */
	start_router = current_entry;
	no_router_entries--;
	edited_config = TRUE;
	router_changed = TRUE;
    
      }
      else if (current_entry->next == (struct router_entry *) NULL)
      { /* end of list */
    
	temp_entry = current_entry;
	current_entry = current_entry->prev;
	free((char *) temp_entry);
	if (current_entry != NULL)  /* Not last item in list */
	  current_entry->next = NULL;
	else
	  start_router = NULL; /* Last item in list, set end pointer to null */
	end_router = current_entry;
	no_router_entries--;
	edited_config = TRUE;
	router_changed = TRUE;
	entry_number--;
    
      }
      else
      {   /* Not begining or end of table */
    
	current_entry->prev->next = current_entry->next;
	current_entry->next->prev = current_entry->prev;
	temp_entry = current_entry;
	current_entry = current_entry->next;
	free((char *) temp_entry);
	no_router_entries--;
	edited_config = TRUE;
	router_changed = TRUE;
      }
    }
    /*}}}  */
    /*{{{  goto entry*/
    else if (((addr_line[0] == 'G') || (addr_line[0] == 'g')) &&
	     (start_router != NULL))
    {
      int line_no;
    
      do
      {
    
	printf("Entry number [1 - 127]: ");
	fgets(addr_line, sizeof(addr_line), stdin);
	tidy_string(addr_line);
	len = sscanf( addr_line, "%d%c", &line_no, &dummy);
    
      } while ((len != 0) && (len != 1));
      if ((len == 1) && (line_no >= 1))
      {
	entry_number = 1;
	current_entry = start_router;
	for(; (current_entry->next != NULL) && (line_no > 1);
				line_no--, current_entry = current_entry->next)
	  entry_number++;  /* move to required item (or end) */
    
      }
    }
    /*}}}  */
    /*{{{  edit entry*/
    else if (((addr_line[0] == 'E') || (addr_line[0] == 'e')) &&
	     (current_entry != NULL))
    {
      BOOL changed;
	  
      changed = edit_router_entry(current_entry);
      if (changed == TRUE)
	router_changed = TRUE;
    
    }
    /*}}}  */
    /*{{{  too many entries*/
    else if (no_router_entries > MAX_ROUTER_ENTRIES)
      printf("No more room for router entries\n");
    /*}}}  */
    /*{{{  new entry*/
    else if ((addr_line[0] == 'D') || (addr_line[0] == 'n'))
    {
    
     struct router_entry *temp_entry;
    
      temp_entry = (struct router_entry *) malloc(sizeof(struct router_entry));
      /*{{{  malloc failed*/
      if (temp_entry == NULL)
      {
	printf("Error: MALLOC FAILED\n");
      }
      /*}}}  */
      /*{{{  otherwise, get data and try to insert after*/
      else
      {
	BOOL ok;
      
	ok = get_router_entry(temp_entry);
      
	if (ok == TRUE)
	{
	  edited_config = TRUE;
	  router_changed = TRUE;
	  if (no_router_entries == 0)
	  {
	    temp_entry->prev = NULL;
	    temp_entry->next = NULL;
	    start_router = temp_entry;
	    end_router = temp_entry;
	    entry_number = 1;
	  }
	  else   /* insert after */
	  {
	    if (current_entry->next != NULL) /* Not at end */
	    {
	      current_entry->next->prev = temp_entry;
	    }
	    else
	    {
	      end_router = temp_entry;
	    }
	    temp_entry->next = current_entry->next;
	    temp_entry->prev = current_entry;
	    current_entry->next = temp_entry;
	    entry_number++;
	  }
      
	  current_entry = temp_entry;
	  no_router_entries++;
	}
	else /* ok != TRUE */
	{
	  printf("Incomplete entry not added to router table\n");
	  free((char *) temp_entry);
	}
      }
      /*}}}  */
    
    }
    /*}}}  */
    /*{{{  insert entry*/
    else if ((addr_line[0] == 'I') || (addr_line[0] == 'i'))
    {
      struct router_entry *temp_entry;
    
      temp_entry = (struct router_entry *) malloc(sizeof(struct router_entry));
      /*{{{  malloc failed*/
      if (temp_entry == NULL)
      {
	 printf("Error: MALLOC FAILED\n");
      }
      /*}}}  */
      /*{{{  otherwise, get data and try to insert before*/
      else
      {
	BOOL ok;
      
	ok = get_router_entry(temp_entry);
      
	if (ok == TRUE)
	{
	  edited_config = TRUE;
	  router_changed = TRUE;
	  if (no_router_entries == 0)
	  {
	    temp_entry->prev = NULL;
	    temp_entry->next = NULL;
	    start_router = temp_entry;
	    end_router = temp_entry;
	    current_entry = temp_entry;
	    entry_number = 1;
	  }
	  else   /* insert before */
	  {
	    if (current_entry->prev != NULL) /* Not at start */
	      current_entry->prev->next = temp_entry;
	    else /* start of list */
	      start_router = temp_entry;
	    temp_entry->next = current_entry;
	    temp_entry->prev = current_entry->prev;
	    current_entry->prev = temp_entry;
	    entry_number++;
	  }
	  no_router_entries++;
	}
	else /* ok != TRUE */
	{
	  printf("Incomplete entry not added to router table\n");
	  free((char *) temp_entry);
	}
      }
      /*}}}  */
    
    }
    /*}}}  */
  }
  
  if ((router_changed) && (no_router_entries == 0))
    init_brouter(&current_config_net[ROUTER_OFFSET]);
  else if (router_changed)
    copy_router_list(&current_config_net[ROUTER_OFFSET], start_router);
  /*}}}  */

  return;
}
/*}}}  */
/*}}}  */

/*{{{  load network*/

void load_network ()
/*
 * Top level routine to load network data from various sources
 */
{
  char filename[MAX_NAME_LEN + 1];   /* extra byte for null char */
  char command[COMMAND_LEN];
  BOOL looping = TRUE;

  printf("Load network data\n");
  while (looping)
  {
    printf("Give network data source (F) [F(ile), N(vram), R(am), Q(uit), ?]: ");
    fgets(command, sizeof(command), stdin);
    switch (command[0])
    {
      case 'N': case 'n':
	current_net_len = read_net_NVRAM(current_config_net);
	edited_config = FALSE;
	looping = FALSE;
	break;
      case 'F': case 'f':  case '\n':
	printf("File name (%s): ", saved_net_filename);
	fgets(filename, MAX_NAME_LEN, stdin);
	tidy_string(filename);
	if (filename[0] == '\0')
	  strcpy(filename, saved_net_filename);
	else
	  strcpy(saved_net_filename, filename);
	current_net_len = read_net_file(current_config_net, filename);
	edited_config = FALSE;
	looping = FALSE;
	break;
      case 'R': case 'r':
	current_net_len = read_net_RAM(current_config_net);
	edited_config = FALSE;
	looping = FALSE;
	break;
      case '?':
	load_help();
	break;
      case 'Q': case 'q':
	looping = FALSE;
	break;
      default:
	looping = FALSE;
	break;
    }
  }

  return;
}


/*}}}  */
/*{{{  load resource*/

void load_resource ()
/*
 * Top level routine to load resource data from various sources
 *
 */
{
  char systemname[MAX_NAME_LEN + 1];   /* extra byte for null char */
  char command[COMMAND_LEN];
  char *environment_ptr;
  BOOL looping = TRUE;

  printf("Load resource data\n");
  while (looping)
  {
    printf("Resource data source (F) [F(ile), N(vram), R(am), Q(uit), ?]: ");
    fgets(command, sizeof(command), stdin);
    switch (command[0])
    {
      case 'N': case 'n':
	current_resource_len = read_resource_NVRAM(current_config_resource);
	looping = FALSE;
	break;
      case 'F': case 'f': case '\n':
	environment_ptr = getenv(RESOURCE_ENVIRONMENT);
	if (environment_ptr != NULL)
	{
	  printf("Resource file is %s\n", environment_ptr);
	  do
	  {
	    memcpy(systemname, (char *) &current_config_net[SYSTEM_NAME_OFFSET + 1],
		    current_config_net[SYSTEM_NAME_OFFSET]);
	    systemname[current_config_net[SYSTEM_NAME_OFFSET]] = '\0';
	    printf("TCPlink system name (%s):", systemname);
	    fgets(systemname, MAX_NAME_LEN, stdin);
	    tidy_string(systemname);
	    if (systemname[0] == '\0')
	    {
	      memcpy(systemname, (char *) &current_config_net[SYSTEM_NAME_OFFSET + 1],
		      current_config_net[SYSTEM_NAME_OFFSET]);
	      systemname[current_config_net[SYSTEM_NAME_OFFSET]] = '\0';
	    }
	  } while (systemname[0] == '\0');
	  current_resource_len = read_resource_file(current_config_resource,
						    systemname);
	}
	else
	{
	  printf("Error: %s %s not defined\n", environment_msg, RESOURCE_ENVIRONMENT);
	}
	looping = FALSE;
	break;
      case 'R': case 'r':
	current_resource_len = read_resource_RAM(current_config_resource);
	looping = FALSE;
	break;
      case '?':
	load_help();
	break;
      case 'Q': case 'q':
	looping = FALSE;
	break;
      default:
	looping = FALSE;
	break;
    }
  }

  return;
}


/*}}}  */

/*{{{  save network*/

void save_network ()
/*
 * Save network data file
 */
{
  FILE *fid;
  int byte_count = CHANGE_OFFSET; /* Set byte count to be at start of IP address */
  char filename[MAX_NAME_LEN], name[MAX_DOMAIN_LEN];
  unsigned int addr[IP_ADDR_LEN];
  unsigned int tos;
  unsigned int linkops_port;


  printf("Save network data\n");
  printf("Filename (%s): ", saved_net_filename);
  fgets(filename, sizeof(filename), stdin);
  tidy_string(filename);
  if (filename[0] == '\0')
    strcpy(filename, saved_net_filename);
  fid = fopen(filename, "w");
  if (fid == NULL)
    printf("Error: unable to open network data file %s for writing\n", filename);
  else
  {
     strcpy(saved_net_filename, filename);
     fprintf(fid, "%s\n", netfile_version);
     fprintf(fid, "-- TCPlink address %s,  ",
		   hex_colon_addr(ETHER_ADDR_LEN, ether_addr));
     fprintf(fid, "version %d.%d\n",
		  (int)(current_config_net[MAJOR_OFFSET] & 0XFF),
		  (int)(current_config_net[MINOR_OFFSET] & 0XFF));

     /*
     * Ip address
     */
     addr_toword(addr, &current_config_net[byte_count], IP_ADDR_LEN);
     fprintf(fid, "%s\n", dec_dot_addr(IP_ADDR_LEN, addr));
     byte_count += IP_ADDR_LEN;
     /*
      * Sub net address
      */
     addr_toword(addr, &current_config_net[byte_count], IP_ADDR_LEN);
     fprintf(fid, "%s\n", dec_dot_addr(IP_ADDR_LEN, addr));
     byte_count += IP_ADDR_LEN;
     /*
      * Broadcast address
      */
     addr_toword(addr, &current_config_net[byte_count], IP_ADDR_LEN);
     fprintf(fid, "%s\n", dec_dot_addr(IP_ADDR_LEN, addr));
     byte_count +=  IP_ADDR_LEN;
     /*
      * name server 1 address
      */
     addr_toword(addr, &current_config_net[byte_count], IP_ADDR_LEN);
     fprintf(fid, "%s\n", dec_dot_addr(IP_ADDR_LEN, addr));
     byte_count += IP_ADDR_LEN;
     /*
      * name server 2 address
      */
     addr_toword(addr, &current_config_net[byte_count], IP_ADDR_LEN);
     fprintf(fid, "%s\n", dec_dot_addr(IP_ADDR_LEN, addr));
     byte_count +=  IP_ADDR_LEN;
     /*
      * linkops port number
      */
     linkops_port = (unsigned int) (current_config_net[byte_count] << 8);
     linkops_port = linkops_port + current_config_net[byte_count + 1];
     fprintf(fid, "%u\n", linkops_port);
     byte_count +=  2;  /* 2 for short int */
     /*
      * Link speed
      */
     fprintf(fid, "%d\n", index_to_speed(current_config_net[byte_count]));
     byte_count +=  2;  /* 1 for link speed value and 1 byte of padding */
     /*
      * Host name
      */
     memcpy(name, (char *) &current_config_net[byte_count + 1], current_config_net[byte_count]);
     name[current_config_net[byte_count]] = '\0';
     fprintf(fid, "%s\n", name);
     /*
      * Domain name
      */
     byte_count = DOMAIN_NAME_OFFSET;
     memcpy(name, (char *) &current_config_net[byte_count + 1], current_config_net[byte_count]);
     name[current_config_net[byte_count]] = '\0';
     fprintf(fid, "%s\n", name);

     /*
      * Now the router table
      */

     byte_count = ROUTER_OFFSET;
     while ((byte_count < NET_DATA_SIZE) &&
	   (((unsigned int)(current_config_net[byte_count] & 0XFF) & END_OF_ROUTER_TABLE) != END_OF_ROUTER_TABLE))
     {

       fprintf(fid, "%u ", (unsigned int) (current_config_net[byte_count] & TYPE_MASK));
       byte_count++;
       addr_toword(addr, &current_config_net[byte_count], IP_ADDR_LEN);
       fprintf(fid, "%s ", dec_dot_addr(IP_ADDR_LEN, addr));
       byte_count += IP_ADDR_LEN;
       tos = (unsigned int) (current_config_net[byte_count] & 0XFF);
       byte_count++;
       tos += ((unsigned int) (current_config_net[byte_count] & 0XFF) << 8);
       byte_count++;
       fprintf(fid, "%u ", tos);
       addr_toword(addr, &current_config_net[byte_count], IP_ADDR_LEN);
       fprintf(fid, "%s\n", dec_dot_addr(IP_ADDR_LEN, addr));
       byte_count += IP_ADDR_LEN;

     }

     if (fclose(fid) != 0)
      printf("Error: closing network data file %s\n", filename);
     else
      edited_config = FALSE;
  }

}

/*}}}  */

/*{{{  show network*/

static void show_network (length, config)
int length;
unsigned char config[];
/*
 * Display given network data
 */
{
  char name[MAX_NAME_LEN], domain[MAX_DOMAIN_LEN];
  int index = 0;   /* Set to start of ethernet address */
  int router_count = 1;
  unsigned int ether_bytes[ETHER_ADDR_LEN];
  unsigned int ip_bytes[IP_ADDR_LEN];
  unsigned int tos;
  unsigned int ip_gate_bytes[IP_ADDR_LEN];
  unsigned short int linkops_port;

  addr_toword(ether_bytes, &config[index], ETHER_ADDR_LEN);
  printf("TCPlink address:     %s\n",
		   hex_colon_addr(ETHER_ADDR_LEN, ether_bytes));
  printf("TCPlink version      %d.%d\n", (int)(current_config_net[MAJOR_OFFSET] & 0XFF),
			      (int)(current_config_net[MINOR_OFFSET] & 0XFF));
  index += CHANGE_OFFSET; /* move past addr and version info */
  addr_toword(ip_bytes, &config[index], IP_ADDR_LEN);
  printf("IP address:          %s\n", dec_dot_addr(IP_ADDR_LEN, ip_bytes));
  index += IP_ADDR_LEN;
  addr_toword(ip_bytes, &config[index], IP_ADDR_LEN);
  printf("Sub net mask:        %s\n", dec_dot_addr(IP_ADDR_LEN, ip_bytes));
  index += IP_ADDR_LEN;
  addr_toword(ip_bytes, &config[index], IP_ADDR_LEN);
  printf("Broadcast address:   %s\n", dec_dot_addr(IP_ADDR_LEN, ip_bytes));
  index += IP_ADDR_LEN;
  addr_toword(ip_bytes, &config[index], IP_ADDR_LEN);
  printf("Name server 1:       %s\n", dec_dot_addr(IP_ADDR_LEN, ip_bytes));
  index += IP_ADDR_LEN;
  addr_toword(ip_bytes, &config[index], IP_ADDR_LEN);
  printf("Name server 2:       %s\n", dec_dot_addr(IP_ADDR_LEN, ip_bytes));
  index += IP_ADDR_LEN;
  linkops_port = (unsigned short int) ((config[index] << 8) + config[index + 1]);
  printf("Linkops port:        %u\n", linkops_port);
  index += 2; /* Skip short int and two bytes of padding */
  printf("Link speed:          %d\n", index_to_speed(current_config_net[index]));
  index += 2; /* Skip link spedd and byte of padding */

  memcpy(name, (char *) &config[index + 1], config[index]);
  name[config[index]] = '\0';
  printf("TCPlink system name: %s\n", name);

  index = DOMAIN_NAME_OFFSET;
  memcpy(domain, (char *) &config[index + 1], config[index]);
  domain[config[index]] = '\0';
  printf("Domain name:         %s\n", domain);

  /* Now for the router table */

  index = ROUTER_OFFSET;
  printf("\nRouter table:\n\n");
  /*{{{  */
  while ((index < NET_DATA_SIZE) &&
	 (((unsigned int)(config[index] & 0XFF) & END_OF_ROUTER_TABLE) != END_OF_ROUTER_TABLE))
  {
    addr_toword(ip_bytes, &config[index + 1], IP_ADDR_LEN);
    tos = (unsigned int) (config[index + 5] & 0XFF);
    tos = tos + ((unsigned int) (config[index + 6] & 0XFF) << 8);
    addr_toword(ip_gate_bytes, &config[index + 7], IP_ADDR_LEN);
  
    printf("%d: ", router_count);
    printf("Type: %s, ", show_type(config[index]));
    if (((int) (config[index] & 0XFF) & DEFAULT_TYPE) != DEFAULT_TYPE)
      printf("Address: %s, ", dec_dot_addr(IP_ADDR_LEN, ip_bytes));
  
    if (((int) (config[index] & 0XFF) & NO_TOS_TYPE) != NO_TOS_TYPE)
      printf("TOS: %u, ", tos);
    printf("Gateway: %s\n", dec_dot_addr(IP_ADDR_LEN, ip_gate_bytes));
  
    index += ROUTER_ENTRY_SIZE;
    router_count++;
  
  }
  /*}}}  */

  if (router_count == 1)
  {
    printf("No router entries\n");
  }

  return;
}

/*}}}  */
/*{{{  display_network*/

void display_network ()
/*
 * Top level routine to display network data from various sources
 */
{
  char filename[MAX_NAME_LEN + 1];   /* extra byte for null char */
  char command[COMMAND_LEN];
  BOOL looping = TRUE;

  printf("Display network data\n");
  while (looping)
  {
    printf("Network data source (C) [C(urrent), F(ile), N(vram), R(am), Q(uit), ?]: ");
    fgets(command, sizeof(command), stdin);
    switch (command[0])
    {
      case 'C': case 'c': case '\n':
	show_network(current_net_len, current_config_net);
	looping = FALSE;
	break;
      case 'N': case 'n':
	display_net_len = read_net_NVRAM(display_config_net);
	show_network(display_net_len, display_config_net);
	looping = FALSE;
	break;
      case 'F': case 'f':
	printf("File name (%s): ", saved_net_filename);
	fgets(filename, MAX_NAME_LEN, stdin);
	tidy_string(filename);
	if (filename[0] == '\0')
	  strcpy(filename, saved_net_filename);
	else
	  strcpy(saved_net_filename, filename);
	display_net_len = read_net_file(display_config_net, filename);
	if (display_net_len != 0)
	  show_network(display_net_len, display_config_net);
	looping = FALSE;
	break;
      case 'R': case 'r':
	display_net_len = read_net_RAM(display_config_net);
	show_network(display_net_len, display_config_net);
	looping = FALSE;
	break;
      case '?':
	load_help();
	break;
      case 'Q': case 'q':
	looping = FALSE;
	break;
      default:
	looping = FALSE;
	break;
    }
  }

  return;
}


/*}}}  */
/*{{{  show resource*/

static void show_resource (length, config)
int length;
unsigned char config[];
/*
 * Display given resource data
 */
{
  char name[MAX_NAME_LEN];
  int index = 0;
  int count = 0;
  int name_len = (int) (config[index + RESOURCE_LEN_O] & 0XFF);


  while ((index < length) && (name_len > 0) && (count < MAX_RESOURCE_ENTRIES))
  {

    printf("Link number: %d  ", (int) (config[index] & 0XFF));
    memcpy(name, (char *) &config[index + RESOURCE_NAME_O], name_len);
    name[name_len] = '\0';
    printf("Name: %s\n", name);
    count ++;
    index += RESOURCE_ENTRY_SIZE;
    if (index < length)
    {
      name_len = (int) (config[index + RESOURCE_LEN_O] & 0XFF);
    }
    else
    {
      name_len = 0;
    }

  }


  return;
}


/*}}}  */
/*{{{  display_resource*/

void display_resource ()
/*
 * Top level routine to display resource data from various sources
 */
{
  char systemname[MAX_NAME_LEN + 1];   /* extra byte for null char */
  char command[COMMAND_LEN];
  char *environment_ptr;
  BOOL looping = TRUE;

  printf("Display resource data\n");
  while (looping)
  {
    printf("Resource data source (C) [C(urrent), F(ile), N(vram), R(am), Q(uit), ?]: ");
    fgets(command, sizeof(command), stdin);
    putchar('\n');
    switch (command[0])
    {
      case 'C': case 'c': case '\n':
	show_resource(current_resource_len, current_config_resource);
	looping = FALSE;
	break;
      case 'N': case 'n':
	display_resource_len = read_resource_NVRAM(display_config_resource);
	show_resource(display_resource_len, display_config_resource);
	looping = FALSE;
	break;
      case 'F': case 'f':
	environment_ptr = getenv(RESOURCE_ENVIRONMENT);
	if (environment_ptr != NULL)
	{
	  printf("Resource file is %s\n", environment_ptr);
	  do
	  {
	    memcpy(systemname, (char *) &current_config_net[SYSTEM_NAME_OFFSET + 1],
		    current_config_net[SYSTEM_NAME_OFFSET]);
	    systemname[current_config_net[SYSTEM_NAME_OFFSET]] = '\0';
	    printf("TCPlink system name (%s):", systemname);
	    fgets(systemname, MAX_NAME_LEN, stdin);
	    tidy_string(systemname);
	    if (systemname[0] == '\0')
	    {
	      memcpy(systemname, (char *) &current_config_net[SYSTEM_NAME_OFFSET + 1],
		      current_config_net[SYSTEM_NAME_OFFSET]);
	      systemname[current_config_net[SYSTEM_NAME_OFFSET]] = '\0';
	    }
	  } while (systemname[0] == '\0');
	  display_resource_len = read_resource_file(display_config_resource,
						    systemname);
	  show_resource(display_resource_len, display_config_resource);
	}
	else
	{
	  printf("Error: %s %s not defined\n", environment_msg, RESOURCE_ENVIRONMENT);
	}
	looping = FALSE;
	break;
      case 'R': case 'r':
	display_resource_len = read_resource_RAM(display_config_resource);
	show_resource(display_resource_len, display_config_resource);
	looping = FALSE;
	break;
      case '?':
	load_help();
	break;
      case 'Q': case 'q':
	looping = FALSE;
	break;
      default:
	looping = FALSE;
	break;
    }
  }

  return;
}


/*}}}  */

/*{{{  write configuration*/

void write_configuration ()
/*
 * Write current configuration to NVRAM
 */
{
  BOOL ok;
  BOOL error = FALSE;
  short int len_to_send = 0;
  int index = 0;
  unsigned short int address = 0;

  printf("Write configuration to NVRAM\n");

  /*
   * First the network data...
   */

  while (index < NET_DATA_SIZE)
  {
    if ((NET_DATA_SIZE - index) >= DATA_LEN)
      len_to_send = DATA_LEN;
    else
      len_to_send = NET_DATA_SIZE - index;

    ok = NVRAM_write(&current_config_net[index], address, len_to_send);
    if (ok == FALSE)
    {
      error = TRUE;
    }
    address += len_to_send;
    index += (int) len_to_send;
  }

  /*
   * ... and then the resource data
   */

  index = 0;
  while (index < RESOURCE_DATA_BLOCK)
  {
    if ((RESOURCE_DATA_BLOCK - index) >= DATA_LEN)
      len_to_send = DATA_LEN;
    else
      len_to_send = RESOURCE_DATA_BLOCK - index;

    ok = NVRAM_write(&current_config_resource[index], address, len_to_send);
    if (ok == FALSE)
    {
      error = TRUE;
    }
    address += len_to_send;
    index += (int) len_to_send;
  }


  if (error == TRUE)
  {
    printf("Warning: write of configuration to NVRAM may have failed\n");
    /* printf("Warning: NVRAM may be corrupted\n"); */
  }
  else
    edited_config = FALSE;

  return;
}
/*}}}  */
/*{{{  verify_configuration*/

void verify_configuration ()
/*
 * Compare current configuration data with that from NVRAM
 */
{
  BOOL error = FALSE;
  int index = 0;
  int router_count = 1;
  int router_error_count = 0;
  int resource_count = 1;
  int resource_error_count = 0;

  printf("Verify configuration\n");

  /*
   * Load data from  NVRAM to display data area
   */

  display_net_len = read_net_NVRAM(display_config_net);
  display_resource_len = read_resource_NVRAM(display_config_resource);

  /*
   * Compare network data area
   */

  /*{{{  basic data*/
  if (memcmp((char *) &display_config_net[index], (char *) &current_config_net[index], 
	      ETHER_ADDR_LEN))
  {
    printf("Ethernet addresses do not match\n");
    error = TRUE;
  }
  index += ETHER_ADDR_LEN + 2;  /* Skip address plus padding */
  
  if (memcmp((char *) &display_config_net[index], (char *) &current_config_net[index], IP_ADDR_LEN))
  {
    printf("TCPlink system IP addresses do not match\n");
    error = TRUE;
  }
  index += IP_ADDR_LEN;  /* Skip address */
  
  if (memcmp((char *) &display_config_net[index], (char *) &current_config_net[index], IP_ADDR_LEN))
  {
    printf("Subnet mask IP addresses do not match\n");
    error = TRUE;
  }
  index += IP_ADDR_LEN;  /* Skip address */
  
  if (memcmp((char *) &display_config_net[index], (char *) &current_config_net[index], IP_ADDR_LEN))
  {
    printf("Broadcast IP addresses do not match\n");
    error = TRUE;
  }
  index += IP_ADDR_LEN;  /* Skip address */
  
  if (memcmp((char *) &display_config_net[index], (char *) &current_config_net[index], IP_ADDR_LEN))
  {
    printf("Primary name server IP addresses do not match\n");
    error = TRUE;
  }
  index += IP_ADDR_LEN;  /* Skip address */
  
  if (memcmp((char *) &display_config_net[index], (char *) &current_config_net[index], IP_ADDR_LEN))
  {
    printf("Secondary name server IP addresses do not match\n");
    error = TRUE;
  }
  index += IP_ADDR_LEN;  /* Skip address */
  
  /* Linkops port test */
  if (memcmp((char *) &display_config_net[index], (char *) &current_config_net[index], 2))
  {
    printf("Linkops port numbers do not match\n");
    error = TRUE;
  }
  index += 2;  /* Skip port number */
  
  /* Link speed test */
  if (display_config_net[index] != current_config_net[index])
  {
    printf("Link speeds do not match\n");
    error = TRUE;
  }
  index += 2;  /* Skip link speed number and one byte of padding*/
  
  if (memcmp((char *) &display_config_net[index], (char *) &current_config_net[index], MAX_NAME_LEN + 1))
  {
    printf("TCPlink system names do not match\n");
    error = TRUE;
  }
  index += MAX_NAME_LEN + 1 ;  /* Skip name and length */
  
  if (memcmp((char *) &display_config_net[index], (char *) &current_config_net[index], MAX_DOMAIN_LEN + 1))
  {
    printf("Network domain names do not match\n");
    error = TRUE;
  }
  index += MAX_DOMAIN_LEN + 1 + 3;  /* Skip domain name, length and padding */
  /*}}}  */

  /*
   * Compare router data area
   */

  /*{{{  router table*/
  while ((index < NET_DATA_SIZE) && (router_error_count < MAX_ERRORS))
  {
    if (memcmp((char *) &display_config_net[index], (char *) &current_config_net[index], ROUTER_ENTRY_SIZE))
    {
      int i;
      BOOL null_entry = TRUE;
  
      for (i = index; i < (index + ROUTER_ENTRY_SIZE); i++)
      {
	if (display_config_net[i] != (unsigned char) 0)
	  null_entry = FALSE;
      }
      if ((null_entry == TRUE) &&
	  ((current_config_net[index] & END_OF_ROUTER_TABLE) == END_OF_ROUTER_TABLE))
	; /* SKIP if display entry is null and current entry is end of table */
      else if (((display_config_net[index] & END_OF_ROUTER_TABLE) == END_OF_ROUTER_TABLE) &&
	       ((current_config_net[index] & END_OF_ROUTER_TABLE) == END_OF_ROUTER_TABLE))
	; /* SKIP if both are some sort of end of table entry */
      else
      {
	printf("Router table entry %d does not match\n", router_count);
	router_error_count++;
      }
    }
  
    index += ROUTER_ENTRY_SIZE;  /* Skip entry */
    router_count++;
  }
  
  
  if (router_error_count >= MAX_ERRORS)
    printf("Too many router table errors\n");
  /*}}}  */


  /*
   * Compare resource data area
   */

  index = 0;
  while ((index < RESOURCE_DATA_SIZE) && (resource_error_count < MAX_ERRORS))
  {
    if (memcmp((char *) &display_config_resource[index],
	       (char *) &current_config_resource[index], RESOURCE_ENTRY_SIZE))
    {
      printf("Resource entry %d does not match\n", resource_count);
      resource_error_count++;
    }
    index += RESOURCE_ENTRY_SIZE;  /* Skip entry */
    resource_count++;
  }


  if (resource_error_count >= MAX_ERRORS)
    printf("Too many resource data errors\n");

  if ((! error) && (router_error_count == 0) && (resource_error_count == 0))
    printf("Configuration verified OK\n");

  return;
}
/*}}}  */
