/* itstat.c */
/* Copyright INMOS Limited 1992 */
/* Version 1.2 */

#ifndef lint
static char    *SCCSid = "@(#)itstat.c v1.2: Copyright INMOS Limited 1992";

#endif				/* lint */

/*
 ******************************************************************************
 *
 * Program:	itstat.c
 *
 * Purpose:	Simple driver program to query INMOS networking products
 *		using Tricklet SNMP software.
 *
 *		Tricklet software courtesy of the DNPAP group,
 *		Technical University Delft, Holland.
 *
 * Author:	Graham W. Hudspith
 *
 * Date:	11/6/92
 *
 ******************************************************************************
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifdef PCNFS
#include <tklib.h>
#include <sys/tk_types.h>
#include <netinet/in.h>
#include <in_addr.h>
#else
#include <sys/types.h>

#ifdef VAX

#ifdef WINTCP

#ifndef NETINET_IN.H
#include <netinet/in.h>
#define NETINET_IN.H
#endif				/* NETINET_IN.H */

#endif				/* WINTCP */

#endif				/* VAX */

#include <arpa/inet.h>
#endif				/* PCNFS */

#define PROGRAM_NAME	"itstat"
#define VERSION_NAME	"1.2 (dated 11/6/92)"

#define F005_OBJID	"1.3.6.1.4.1.281.1.1."
#define F009_OBJID	"1.3.6.1.4.1.281.1.2."

#ifdef __sun
/*
 * If compiled on a Sun with gcc (v2.2.2 or greater), define either
 * sun3 or sun4.
 */

#ifdef __mc68000
#define sun3
#endif				/* __mc68000 */

#ifdef __sparc
#define sun4
#endif				/* __sparc */

#endif				/* __sun */

#ifdef sun3
#define HOST		"Sun3/SunOS4"
#endif				/* sun3 */

#ifdef sun4
#define HOST		"Sun4/SunOS4"
#endif				/* sun4 */

#ifdef MSDOS

#ifdef NEC

#ifdef PCNFS
#define HOST		"NEC-PC/MS-DOS PC-NFS"
#endif				/* PCNFS */

#ifdef PCTCP
#define HOST		"NEC-PC/MS-DOS PC/TCP"
#endif				/* PCTCP */

#else

#ifdef PCNFS
#define HOST		"IBM-PC/MS-DOS PC-NFS"
#endif				/* PCNFS */

#ifdef PCTCP
#define HOST		"IBM-PC/MS-DOS PC/TCP"
#endif				/* PCTCP */

#endif				/* NEC */

#endif				/* MSDOS */

#ifdef VMS

#ifdef DECTCP
#define HOST		"VAX/VMS VMS/Ultrix Connection"
#endif				/* DECTCP */

#ifdef WINTCP
#define HOST		"VAX/VMS WIN/TCP"
#endif				/* WINTCP */

#endif				/* VMS */

#include "getopt.h"

#include "dnpap.h"
#include "snmp.h"
#include "parse.h"
#include "mib.h"

#include "mibdefs.h"

void            print_help (void);
void            ImsFullMib (void);
int             do_key (mgt_sck *, const char *const, const unsigned long,
			const int, const char, const char *const, const int);
void            do_query (mgt_sck *, const char *const, const unsigned long,
			  const int, const int, char *, char *, const int);
void            config_scr (mgt_sck *, const char *const,
			    const unsigned long, const int);
void            help_scr (const char *const, const int);
void            history_scr (mgt_sck *, const char *const,
			     const unsigned long, const int);
void            icmp_scr (mgt_sck *, const char *const, const unsigned long,
			  const int);
void            ip_scr (mgt_sck *, const char *const, const unsigned long,
			const int);
void            memory_scr (mgt_sck *, const char *const,
			    const unsigned long, const int);
void            snmp_scr (mgt_sck *, const char *const, const unsigned long,
			  const int);
void            subsys_scr (mgt_sck *, const char *const,
			    const unsigned long, const int);
void            tcp_data_scr (mgt_sck *, const char *const,
			      const unsigned long, const int);
void            tcp_misc_scr (mgt_sck *, const char *const,
			      const unsigned long, const int);
void            udp_scr (mgt_sck *, const char *const, const unsigned long,
			 const int);

int
main (int argc,
      char *argv      [])
{
    int             key, cflag, pflag, rflag, tflag, errflag;
    char            community[256];
    int             port, retries;
    unsigned long   timeout;
    char            host[256], args[256];
    int             argslen, quit;
    mgt_sck        *handle;

    strcpy (community, "public");
    port = 161;
    retries = 3;
    timeout = 5000000L;		/* 5 second timeout */

    cflag = 0;			/* has user passed in community arg? */
    pflag = 0;			/* has user passed in port arg? */
    rflag = 0;			/* has user passed in retries arg? */
    tflag = 0;			/* has user passed in timeout arg? */
    errflag = 0;

    while ((key = getopt (argc, argv, "c:p:r:t:")) != -1)
    {
	switch (key)
	{
	case 'c':
	    strcpy (community, optarg);
	    if (cflag || strlen (community) <= 0)
	    {
		errflag++;
	    }
	    else
	    {
		cflag = 1;
	    }
	    break;

	case 'p':
	    port = atoi (optarg);
	    if (pflag || port <= 0)
	    {
		errflag++;
	    }
	    else
	    {
		pflag = 1;
	    }
	    break;

	case 'r':
	    retries = atoi (optarg);
	    if (rflag || retries <= 0)
	    {
		errflag++;
	    }
	    else
	    {
		rflag = 1;
	    }
	    break;

	case 't':
	    timeout = atol (optarg);
	    if (tflag || timeout == 0)
	    {
		errflag++;
	    }
	    else
	    {
		tflag = 1;
	    }
	    break;

	case '?':
	default:
	    errflag++;
	}

	if (errflag)
	    break;
    }

    if (errflag || (argc - optind) < 1 || (argc - optind) > 2)
    {
	print_help ();
	exit (1);
    }

    strcpy (host, argv[optind++]);

    if ((argc - optind) == 1)
    {
	strcpy (args, argv[optind++]);
	argslen = strlen (args);
    }
    else
    {
	args[0] = '\0';
	argslen = 0;
    }

    if ((handle = MgtOpn (htons (port), MgtAdr (host))) == NULL)
    {
	fprintf (stderr, "Opening SNMP port to %s failed\n", host);
	exit (1);
    }

    if (MibInit () != 0)
    {
	fprintf (stderr, "Mib Initialisation failed\n");
	exit (1);
    }

    /*
     * A true 'tricklet' program would call MibFullMib () here, which would
     * read in the mibs from a set of files (allowing the user to edit those
     * files). For portability, we specify the mibs in a large array of strings
     * (see mibdefs.h). The downside is that this leaves us less configurable.
     */
    ImsFullMib ();

    /*
     * User can specify 'keys' of screens needed on the command line, or go
     * into interactive loop.
     */
    if (argslen > 0)
    {
	for (key = 0; key < argslen; key++)
	{
	    (void) do_key (handle, community, timeout, retries,
			   args[key], PROGRAM_NAME, 0);
	}
    }
    else
    {
	help_scr (PROGRAM_NAME, 1);

	printf ("\n%s> ", PROGRAM_NAME);
	fflush (stdout);

	quit = 0;
	while (!quit && gets (args) != NULL)
	{
	    if (strlen (args) > 0)
	    {
		quit = do_key (handle, community, timeout, retries,
			       args[0], PROGRAM_NAME, 1);
	    }

	    if (!quit)
	    {
		printf ("\n%s> ", PROGRAM_NAME);
		fflush (stdout);
	    }
	}

	printf ("\n");
    }

    MgtCls (handle);

    exit (0);
}

void
print_help (void)
{
    printf ("\n%s : INMOS TCPlink statistics\n", PROGRAM_NAME);
    printf ("%s Version %s.\n", HOST, VERSION_NAME);
    printf ("Copyright INMOS Limited 1992.\n\n");
    printf ("Usage:  %s { -options } TCPlink_box { arguments }\n\n",
	    PROGRAM_NAME);
    printf ("Options: \n");
    printf (" -c community    SNMP community string\n");
    printf (" -p port         SNMP port number\n");
    printf (" -r retries      number of retries of each SNMP request\n");
    printf (" -t timeout      SNMP retry timeout (microseconds) \n\n");
    printf ("Arguments (any combination of): \n");
    printf (" c ICMP statistics               h subsystem history statistics\n");
    printf (" i IP statistics                 m memory statistics\n");
    printf (" n SNMP statistics               p TCP statistics - connection & misc\n");
    printf (" s subsystem information         t TCP statistics - packets sent & received\n");
    printf (" u UDP statistics                v version & configuration information\n\n");
}

int
do_key (mgt_sck * handle,
	const char *const community,
	const unsigned long timeout,
	const int retries,
	const char key,
	const char *const argv0,
	const int interactive)
{
    int             quit = 0;

    switch (key)
    {
    case 'c':
    case 'C':
	icmp_scr (handle, community, timeout, retries);
	break;

    case 'h':
    case 'H':
	history_scr (handle, community, timeout, retries);
	break;

    case 'i':
    case 'I':
	ip_scr (handle, community, timeout, retries);
	break;

    case 'm':
    case 'M':
	memory_scr (handle, community, timeout, retries);
	break;

    case 'n':
    case 'N':
	snmp_scr (handle, community, timeout, retries);
	break;

    case 'p':
    case 'P':
	tcp_misc_scr (handle, community, timeout, retries);
	break;

    case 's':
    case 'S':
	subsys_scr (handle, community, timeout, retries);
	break;

    case 't':
    case 'T':
	tcp_data_scr (handle, community, timeout, retries);
	break;

    case 'u':
    case 'U':
	udp_scr (handle, community, timeout, retries);
	break;

    case 'v':
    case 'V':
	config_scr (handle, community, timeout, retries);
	break;

    case 'q':
    case 'Q':
	quit = 1;
	if (interactive)
	    break;
	/* WARNING - FALL THROUGH */

    case '?':
    default:
	help_scr (argv0, interactive);
	break;
    }

    return (quit);
}

static char    *icmp_vars[] =
{
    "icmpInMsgs",
    "icmpInErrors",
    "icmpInDestUnreachs",
    "icmpInTimeExcds",
    "icmpInSrcQuenchs",
    "icmpInRedirects",
    "icmpInEchos",
    "icmpInEchoReps",
    "icmpOutMsgs",
    "icmpOutErrors",
    "icmpOutDestUnreachs",
    "icmpOutTimeExcds",
    "icmpOutSrcQuenchs",
    "icmpOutEchos",
    "icmpOutEchoReps",
};

static char    *icmp_strs[] =
{
    "Received packets:",
    "Received packets discarded:",
    "Received unreachable:",
    "Received time exceeded:",
    "Received source quench:",
    "Received route redirect:",
    "Received echos:",
    "Received echo replies:",
    "Transmit packet requests:",
    "Transmitted packets discarded:",
    "Transmitted unreachable:",
    "Transmitted time exceeded:",
    "Transmitted source quench:",
    "Transmitted echos:",
    "Transmitted echo replies:",
};

#define NUM_ICMP_VARS (sizeof (icmp_vars) / sizeof (char *))

void
icmp_scr (mgt_sck * handle,
	  const char *const community,
	  const unsigned long timeout,
	  const int retries)
{
    char            reply[256], *replyP;
    int             index;

    printf ("\nICMP statistics:\n");

    for (index = 0; index < NUM_ICMP_VARS; index++)
    {
	do_query (handle, community, timeout, retries, PKT_GETRQS,
		  icmp_vars[index], reply, sizeof (reply));
	if ((replyP = strstr (reply, "=")) == NULL)
	    replyP = reply;
	else
	    replyP++;
	printf (" %-32s%10s\n", icmp_strs[index], replyP);
    }
}

void
history_scr (mgt_sck * handle,
	     const char *const community,
	     const unsigned long timeout,
	     const int retries)
{
    char            query[256], reply[256], *replyP;
    int             number, index, days, hours, mins, secs;
    long            elapsed;

    printf ("\nSubsystem History Information:\n");
    printf (" Subsys     Use count             Elapsed time\n");

    do_query (handle, community, timeout, retries, PKT_GETRQS,
	      "imsSubsysNumber", reply, sizeof (reply));
    if ((replyP = strstr (reply, "=")) == NULL)
	replyP = reply;
    else
	replyP++;
    number = atoi (replyP);

    for (index = 1; index <= number; index++)
    {
	sprintf (query, "imsSubsysUses[%d]", index);
	do_query (handle, community, timeout, retries, PKT_GETRQS,
		  query, reply, sizeof (reply));
	if ((replyP = strstr (reply, "=")) == NULL)
	    replyP = reply;
	else
	    replyP++;

	printf ("%5d     %9s          ", index - 1, replyP);

	sprintf (query, "imsSubsysTime[%d]", index);
	do_query (handle, community, timeout, retries, PKT_GETRQS,
		  query, reply, sizeof (reply));
	if ((replyP = strstr (reply, "=")) == NULL)
	    replyP = reply;
	else
	    replyP++;
	elapsed = atol (replyP) / 100;
	secs = elapsed % 60;
	elapsed /= 60;
	mins = elapsed % 60;
	elapsed /= 60;
	hours = elapsed % 24;
	elapsed /= 24;
	days = elapsed;

	printf ("%d %s, ", days, (days == 1 ? "day" : "days"));
	printf ("%d %s, ", hours, (hours == 1 ? "hour" : "hours"));
	printf ("%d %s, ", mins, (mins == 1 ? "min" : "mins"));
	printf ("%d %s\n", secs, (secs == 1 ? "sec" : "secs"));
    }
}

static char    *ip_vars[] =
{
    "ipInReceives",
    "ipInHdrErrors",
    "ipInAddrErrors",
    "ipInDiscards",
    "ipInDelivers",
    "ipOutRequests",
    "ipOutNoRoutes",
    "ipOutDiscards",
    "ipReasmReqds",
    "ipReasmOKs",
    "ipReasmFails",
    "ipFragOKs",
    "ipFragFails",
    "ipFragCreates",
};

static char    *ip_strs[] =
{
    "Received packets:",
    "Received packet header errors:",
    "Received packet address errors:",
    "Received packets discarded (other error):",
    "Received packets delivered to transport:",
    "Transmit packet requests:",
    "Transmit packets discarded (no route):",
    "Transmit packets discarded (other error):",
    "Received fragments:",
    "Received fragments reassembled:",
    "Received fragments discarded:",
    "Transmit packets fragmented:",
    "Transmit fragments discarded:",
    "Transmit fragments created:",
};

#define NUM_IP_VARS (sizeof (ip_vars) / sizeof (char *))

void
ip_scr (mgt_sck * handle,
	const char *const community,
	const unsigned long timeout,
	const int retries)
{
    char            reply[256], *replyP;
    int             index;

    printf ("\nIP statistics:\n");

    for (index = 0; index < NUM_IP_VARS; index++)
    {
	do_query (handle, community, timeout, retries, PKT_GETRQS,
		  ip_vars[index], reply, sizeof (reply));
	if ((replyP = strstr (reply, "=")) == NULL)
	    replyP = reply;
	else
	    replyP++;
	printf (" %-42s%11s\n", ip_strs[index], replyP);
    }
}

#define BUFSIZE_PREFIX	"imsMemBufIndex"

static char    *mem_vars[] =
{
    "imsMemTotalMemory",
    "imsMemTotalBuffers",
    "imsMemBuffersInUse",
};

static char    *mem_strs[] =
{
    "Total memory (bytes):",
    "Total buffers:",
    "Buffers in use:",
};

#define NUM_MEM_VARS (sizeof (mem_vars) / sizeof (char *))

void
memory_scr (mgt_sck * handle,
	    const char *const community,
	    const unsigned long timeout,
	    const int retries)
{
    char            query[256], reply[256], *replyP;
    int             index, bufsize, buftotal, inuse, failures;

    printf ("\nMemory buffer statistics:\n\n");

    for (index = 0; index < NUM_MEM_VARS; index++)
    {
	do_query (handle, community, timeout, retries, PKT_GETRQS,
		  mem_vars[index], reply, sizeof (reply));
	if ((replyP = strstr (reply, "=")) == NULL)
	    replyP = reply;
	else
	    replyP++;
	printf (" %-22s%10s\n", mem_strs[index], replyP);
    }

    printf ("\n Buffer Size     Total     In Use       Failures\n");
    printf ("   (bytes)     (buffers)  (buffers)\n");

    /*
     * Don't know how many buffer sizes there are (well, I do but that's
     * cheating), so have do get-nexts until run out of buffer sizes.
     */
    do_query (handle, community, timeout, retries, PKT_NXTRQS,
	      "imsMemBuffersInUse[0]", reply, sizeof (reply));
    while (strncmp (reply, BUFSIZE_PREFIX, strlen (BUFSIZE_PREFIX)) == 0)
    {
	if ((replyP = strstr (reply, "=")) == NULL)
	    replyP = reply;
	else
	    replyP++;
	index = atoi (replyP);

	sprintf (query, "imsMemBufSize[%d]", index);
	do_query (handle, community, timeout, retries, PKT_GETRQS,
		  query, reply, sizeof (reply));
	if ((replyP = strstr (reply, "=")) == NULL)
	    replyP = reply;
	else
	    replyP++;
	bufsize = atoi (replyP);

	sprintf (query, "imsMemBufTotal[%d]", index);
	do_query (handle, community, timeout, retries, PKT_GETRQS,
		  query, reply, sizeof (reply));
	if ((replyP = strstr (reply, "=")) == NULL)
	    replyP = reply;
	else
	    replyP++;
	buftotal = atoi (replyP);

	sprintf (query, "imsMemBufInUse[%d]", index);
	do_query (handle, community, timeout, retries, PKT_GETRQS,
		  query, reply, sizeof (reply));
	if ((replyP = strstr (reply, "=")) == NULL)
	    replyP = reply;
	else
	    replyP++;
	inuse = atoi (replyP);

	sprintf (query, "imsMemBufFailures[%d]", index);
	do_query (handle, community, timeout, retries, PKT_GETRQS,
		  query, reply, sizeof (reply));
	if ((replyP = strstr (reply, "=")) == NULL)
	    replyP = reply;
	else
	    replyP++;
	failures = atoi (replyP);

	printf ("   %7d      %5d      %4d (%3d%%)    %4d\n",
		bufsize, buftotal, inuse,
		(int) ((((float) inuse * 100.0) / (float) buftotal) + 0.5),
		failures);

	sprintf (query, "%s[%d]", BUFSIZE_PREFIX, index);
	do_query (handle, community, timeout, retries, PKT_NXTRQS,
		  query, reply, sizeof (reply));
    }
}

static char    *trap_modes[] =
{
    "  ???", " True", "False"
};

static char    *snmp_vars[] =
{
    "snmpInPkts",
    "snmpOutPkts",
    "snmpInBadVersions",
    "snmpInBadCommunityNames",
    "snmpInBadCommunityUses",
    "snmpInASNParseErrs",
    "snmpInReadOnlys",
    "snmpInTooBigs",
    "snmpOutTooBigs",
    "snmpInNoSuchNames",
    "snmpOutNoSuchNames",
    "snmpInBadValues",
    "snmpOutBadValues",
    "snmpInGenErrs",
    "snmpOutGenErrs",
    "snmpInGetRequests",
    "snmpOutGetRequests",
    "snmpInGetNexts",
    "snmpOutGetNexts",
    "snmpInSetRequests",
    "snmpOutSetRequests",
    "snmpInGetResponses",
    "snmpOutGetResponses",
    "snmpInTraps",
    "snmpOutTraps",
};

static char    *snmp_strs[] =
{
    "In packets:",
    "Out packets:",
    "Had bad version:",
    "Had bad community name:",
    "Had bad community use:",
    "Had ASN parse errors:",
    "Error = read only:",
    "Error = too big:",
    "Error = too big:",
    "Error = no such name:",
    "Error = no such name:",
    "Error = bad value:",
    "Error = bad value:",
    "Error = general:",
    "Error = general:",
    "Get-Requests:",
    "Get-Requests:",
    "Get-Nexts:",
    "Get-Nexts:",
    "Set-Requests:",
    "Set-Requests:",
    "Get-Responses:",
    "Get-Responses:",
    "Traps:",
    "Traps:",
};

void
snmp_scr (mgt_sck * handle,
	  const char *const community,
	  const unsigned long timeout,
	  const int retries)
{
    char            reply[256], *replyP;
    int             index, value;

    printf ("\nSNMP statistics:\n");

    do_query (handle, community, timeout, retries, PKT_GETRQS,
	      snmp_vars[0], reply, sizeof (reply));
    if ((replyP = strstr (reply, "=")) == NULL)
	replyP = reply;
    else
	replyP++;
    printf (" %-27s%11s", snmp_strs[0], replyP);

    do_query (handle, community, timeout, retries, PKT_GETRQS,
	      snmp_vars[1], reply, sizeof (reply));
    if ((replyP = strstr (reply, "=")) == NULL)
	replyP = reply;
    else
	replyP++;
    printf (" %-27s%11s\n", snmp_strs[1], replyP);

    for (index = 2; index < (2 + 5); index++)
    {
	do_query (handle, community, timeout, retries, PKT_GETRQS,
		  snmp_vars[index], reply, sizeof (reply));
	if ((replyP = strstr (reply, "=")) == NULL)
	    replyP = reply;
	else
	    replyP++;
	printf ("   %-25s%11s\n", snmp_strs[index], replyP);
    }

    for (index = 3; index < (3 + 9); index++)
    {
	do_query (handle, community, timeout, retries, PKT_GETRQS,
		  snmp_vars[index * 2 + 1], reply, sizeof (reply));
	if ((replyP = strstr (reply, "=")) == NULL)
	    replyP = reply;
	else
	    replyP++;
	printf ("   %-25s%11s", snmp_strs[index * 2 + 1], replyP);

	do_query (handle, community, timeout, retries, PKT_GETRQS,
		  snmp_vars[index * 2 + 2], reply, sizeof (reply));
	if ((replyP = strstr (reply, "=")) == NULL)
	    replyP = reply;
	else
	    replyP++;
	printf ("   %-25s%11s\n", snmp_strs[index * 2 + 2], replyP);
    }

    printf ("\n");

    do_query (handle, community, timeout, retries, PKT_GETRQS,
	      "snmpInTotalReqVars", reply, sizeof (reply));
    if ((replyP = strstr (reply, "=")) == NULL)
	replyP = reply;
    else
	replyP++;
    printf (" Total requested variables: %11s\n", replyP);

    do_query (handle, community, timeout, retries, PKT_GETRQS,
	      "snmpInTotalSetVars", reply, sizeof (reply));
    if ((replyP = strstr (reply, "=")) == NULL)
	replyP = reply;
    else
	replyP++;
    printf (" Total set variables:       %11s", replyP);

    do_query (handle, community, timeout, retries, PKT_GETRQS,
	      "snmpEnableAuthenTraps", reply, sizeof (reply));
    if ((replyP = strstr (reply, "=")) == NULL)
	replyP = reply;
    else
	replyP++;
    value = atoi (replyP);
    if (value < 1 || value >= (sizeof (trap_modes) / sizeof (char *)))
    {
	value = 0;
    }

    printf (" Authentication Traps Enabled:    %s\n", trap_modes[value]);
}

static char    *subsys_status[] =
{
    "UNKNOWN", "NOT IN USE", "IN USE", "ERROR FLAG SET",
    "LINK FAILED", "Reserved by configuration"
};

void
subsys_scr (mgt_sck * handle,
	    const char *const community,
	    const unsigned long timeout,
	    const int retries)
{
    char            query[256], reply[256], *replyP;
    int             number, index, status;
    char            resource[256];

    do_query (handle, community, timeout, retries, PKT_GETRQS,
	      "imsSubsysNumber", reply, sizeof (reply));
    if ((replyP = strstr (reply, "=")) == NULL)
	replyP = reply;
    else
	replyP++;
    number = atoi (replyP);

    if (number == 1)
    {
	do_query (handle, community, timeout, retries, PKT_GETRQS,
		  "imsSubsysStatus[1]", reply, sizeof (reply));
	if ((replyP = strstr (reply, "=")) == NULL)
	    replyP = reply;
	else
	    replyP++;
	status = atoi (replyP);
	if (status < 1 || status > (sizeof (subsys_status) / sizeof (char *)))
	{
	    status = 1;
	}

	printf ("\nSubsystem:  %s\n", subsys_status[status - 1]);
    }
    else
    {
	printf ("\nSubsystem Information:\n");
	printf (" Subsystem   Resource name matched            Client\n");

	for (index = 1; index <= number; index++)
	{
	    printf ("%6d       ", index - 1);

	    sprintf (query, "imsSubsysStatus[%d]", index);
	    do_query (handle, community, timeout, retries, PKT_GETRQS,
		      query, reply, sizeof (reply));
	    if ((replyP = strstr (reply, "=")) == NULL)
		replyP = reply;
	    else
		replyP++;
	    status = atoi (replyP);
	    if (status < 1 ||
		    status > (sizeof (subsys_status) / sizeof (char *)))
	    {
		status = 1;
	    }

	    if (status == 3)
	    {
		sprintf (query, "imsSubsysResource[%d]", index);
		do_query (handle, community, timeout, retries, PKT_GETRQS,
			  query, reply, sizeof (reply));
		if ((replyP = strstr (reply, "=")) == NULL)
		    replyP = reply;
		else
		    replyP++;
		strcpy (resource, replyP);

		sprintf (query, "imsSubsysClient[%d]", index);
		do_query (handle, community, timeout, retries, PKT_GETRQS,
			  query, reply, sizeof (reply));
		if ((replyP = strstr (reply, "=")) == NULL)
		    replyP = reply;
		else
		    replyP++;

		printf ("%-32s %s\n", resource, replyP);
	    }
	    else if (status == 6)
	    {
		sprintf (query, "imsSubsysResource[%d]", index);
		do_query (handle, community, timeout, retries, PKT_GETRQS,
			  query, reply, sizeof (reply));
		if ((replyP = strstr (reply, "=")) == NULL)
		    replyP = reply;
		else
		    replyP++;
		printf ("%-32s %s\n", replyP, subsys_status[status - 1]);
	    }
	    else
	    {
		printf ("===%s===\n", subsys_status[status - 1]);
	    }
	}
    }
}

static char    *tcp_data_vars[] =
{
    "tcpInSegs",
    "tcpOutSegs",
    "tcpRetransSegs",
};

static char    *tcp_data_strs[] =
{
    "Received packets:",
    "Transmitted packets:",
    "  Data pkts retxd:",
};

#define NUM_TCP_DATA_VARS (sizeof (tcp_data_vars) / sizeof (char *))

void
tcp_data_scr (mgt_sck * handle,
	      const char *const community,
	      const unsigned long timeout,
	      const int retries)
{
    char            reply[256], *replyP;
    int             index;

    printf ("\nTCP statistics (packets sent and received):\n");

    for (index = 0; index < NUM_TCP_DATA_VARS; index++)
    {
	do_query (handle, community, timeout, retries, PKT_GETRQS,
		  tcp_data_vars[index], reply, sizeof (reply));
	if ((replyP = strstr (reply, "=")) == NULL)
	    replyP = reply;
	else
	    replyP++;
	printf (" %-31s%10s\n", tcp_data_strs[index], replyP);
    }
}

static char    *tcp_misc_vars[] =
{
    "tcpCurrEstab",
    "tcpPassiveOpens",
    "tcpActiveOpens",
    "tcpAttemptFails",
};

static char    *tcp_misc_strs[] =
{
    "Current no. of connections:",
    "Connection requests:",
    "Connection accepts:",
    "Embryonic conn. dropped:",
};

#define NUM_TCP_MISC_VARS (sizeof (tcp_misc_vars) / sizeof (char *))

void
tcp_misc_scr (mgt_sck * handle,
	      const char *const community,
	      const unsigned long timeout,
	      const int retries)
{
    char            reply[256], *replyP;
    int             index;

    printf ("\nTCP statistics (connections etc.):\n");

    for (index = 0; index < NUM_TCP_MISC_VARS; index++)
    {
	do_query (handle, community, timeout, retries, PKT_GETRQS,
		  tcp_misc_vars[index], reply, sizeof (reply));
	if ((replyP = strstr (reply, "=")) == NULL)
	    replyP = reply;
	else
	    replyP++;
	printf (" %-30s%10s\n", tcp_misc_strs[index], replyP);
    }
}

static char    *udp_vars[] =
{
    "udpInDatagrams",
    "udpOutDatagrams",
    "udpInErrors",
};

static char    *udp_strs[] =
{
    "Received packets:",
    "Transmitted packets:",
    "Received packet errors:",
};

#define NUM_UDP_VARS (sizeof (udp_vars) / sizeof (char *))

void
udp_scr (mgt_sck * handle,
	 const char *const community,
	 const unsigned long timeout,
	 const int retries)
{
    char            reply[256], *replyP;
    int             index;

    printf ("\nUDP statistics:\n");

    for (index = 0; index < NUM_UDP_VARS; index++)
    {
	do_query (handle, community, timeout, retries, PKT_GETRQS,
		  udp_vars[index], reply, sizeof (reply));
	if ((replyP = strstr (reply, "=")) == NULL)
	    replyP = reply;
	else
	    replyP++;
	printf (" %-26s%10s\n", udp_strs[index], replyP);
    }
}

static char    *speeds[] =
{
    "unknown", "20 Mbits/sec", "10 Mbits/sec", "5 Mbits/sec"
};

void
config_scr (mgt_sck * handle,
	    const char *const community,
	    const unsigned long timeout,
	    const int retries)
{
    char            query[256], reply[256], *replyP;
    char            version[256], *versionP;
    long            uptime;
    int             padding, days, hours, mins, secs;

    /*
     * Declare some string holders for IP addresses. Make big enough to hold
     * 'xxx.xxx.xxx.xxx\0'.
     */
    char            ip_addr[16], subnet_mask[16];
    int             bcast_mask;
    struct in_addr  bcast_addr;
    unsigned char  *bcast_byteP;

    int             linkops_port, speed;
    int             tftp_config, update_nvram, lookup_names;

    /*
     * Read system description string into 'version' and make 'versionP' point
     * to end.
     */
    do_query (handle, community, timeout, retries, PKT_GETRQS,
	      "sysDescr", reply, sizeof (reply));
    if ((replyP = strstr (reply, "=")) == NULL)
	replyP = reply;
    else
	replyP++;
    /*
     * Move forward over first " if present.
     */
    if (*replyP == '"')
	replyP++;

    strcpy (version, replyP);
    if ((versionP = version + strlen (version)) > version)
    {
	/*
	 * Move back over last " if present.
	 */
	if (versionP > version && *(versionP - 1) == '"')
	{
	    versionP--;
	    *versionP = '\0';
	}

	/*
	 * Read system object identifier string and try to extract version
	 * numbers from it. This information is then appended to end of
	 * 'version'.
	 */
	do_query (handle, community, timeout, retries, PKT_GETRQS,
		  "sysObjectID", reply, sizeof (reply));
	if ((replyP = strstr (reply, F005_OBJID)) != NULL)
	{
	    replyP += strlen (F005_OBJID);
	    sprintf (versionP, " v%s", replyP);
	}
	else if ((replyP = strstr (reply, F009_OBJID)) != NULL)
	{
	    replyP += strlen (F009_OBJID);
	    sprintf (versionP, " v%s", replyP);
	}
    }

    /*
     * Finally, print out 'version'.
     */
    padding = (80 - strlen (version)) / 2;
    printf ("\n%*s%s\n\n", padding, "", version);

    do_query (handle, community, timeout, retries, PKT_GETRQS,
	      "sysUpTime", reply, sizeof (reply));
    if ((replyP = strstr (reply, "=")) == NULL)
	replyP = reply;
    else
	replyP++;
    uptime = atol (replyP) / 100;
    secs = uptime % 60;
    uptime /= 60;
    mins = uptime % 60;
    uptime /= 60;
    hours = uptime % 24;
    uptime /= 24;
    days = uptime;
    printf ("Uptime:              ");
    printf ("%d %s, ", days, (days == 1 ? "day" : "days"));
    printf ("%d %s, ", hours, (hours == 1 ? "hour" : "hours"));
    printf ("%d %s, ", mins, (mins == 1 ? "min" : "mins"));
    printf ("%d %s\n\n", secs, (secs == 1 ? "sec" : "secs"));

    do_query (handle, community, timeout, retries, PKT_GETRQS,
	      "ifPhysAddress[1]", reply, sizeof (reply));
    if ((replyP = strstr (reply, "=")) == NULL)
	replyP = reply;
    else
	replyP++;
    printf ("Ethernet address:    %s\n", replyP);

    do_query (handle, community, timeout, retries, PKT_NXTRQS,
	      "ipFragCreates[0]", reply, sizeof (reply));
    strcpy (query, reply);
    do_query (handle, community, timeout, retries, PKT_NXTRQS,
	      query, reply, sizeof (reply));
    strcpy (query, reply);
    do_query (handle, community, timeout, retries, PKT_GETRQS,
	      query, reply, sizeof (reply));
    if ((replyP = strstr (reply, "=")) == NULL)
	replyP = reply;
    else
	replyP++;
    printf ("IP address:          %s\n", replyP);

    strcpy (ip_addr, replyP);

    do_query (handle, community, timeout, retries, PKT_GETRQS,
	      "sysName", reply, sizeof (reply));
    if ((replyP = strstr (reply, "=")) == NULL)
	replyP = reply;
    else
	replyP++;
    printf ("Name:                %s\n", replyP);

    do_query (handle, community, timeout, retries, PKT_GETRQS,
	      "imsConfigDomainName", reply, sizeof (reply));
    if ((replyP = strstr (reply, "=")) == NULL)
	replyP = reply;
    else
	replyP++;
    printf ("Domain:              %s\n", replyP);

    sprintf (query, "ipAdEntNetMask[%s]", ip_addr);
    do_query (handle, community, timeout, retries, PKT_GETRQS,
	      query, reply, sizeof (reply));
    if ((replyP = strstr (reply, "=")) == NULL)
	replyP = reply;
    else
	replyP++;
    printf ("Subnet mask:         %s\n", replyP);

    strcpy (subnet_mask, replyP);

    sprintf (query, "ipAdEntBcastAddr[%s]", ip_addr);
    do_query (handle, community, timeout, retries, PKT_GETRQS,
	      query, reply, sizeof (reply));
    if ((replyP = strstr (reply, "=")) == NULL)
	replyP = reply;
    else
	replyP++;
    bcast_mask = atoi (replyP);

    bcast_addr.s_addr = inet_addr (ip_addr) & inet_addr (subnet_mask);

    /*
     * Now comes the horrible part ... If bcast_mask is 1, have to convert all
     * '0' parts of bcast_addr into '255'. (e.g. 138.198.0.0 becomes
     * 138.198.255.255).
     */
    if (bcast_mask == 1)
    {
	bcast_byteP = (unsigned char *) &(bcast_addr.s_addr);

	/*
	 * Don't need bcast_mask anymore, so reuse.
	 */
	for (bcast_mask = 0; bcast_mask < sizeof (unsigned long);

		bcast_mask++)
	{
	    if (*bcast_byteP == (unsigned char) 0)
		*bcast_byteP = (unsigned char) 255;
	    bcast_byteP++;
	}
    }

    printf ("Broadcast address:   %s\n", inet_ntoa (bcast_addr));

    do_query (handle, community, timeout, retries, PKT_GETRQS,
	      "imsConfigNameServer1", reply, sizeof (reply));
    if ((replyP = strstr (reply, "=")) == NULL)
	replyP = reply;
    else
	replyP++;
    printf ("Name server 1:       %s\n", replyP);

    do_query (handle, community, timeout, retries, PKT_GETRQS,
	      "imsConfigNameServer2", reply, sizeof (reply));
    if ((replyP = strstr (reply, "=")) == NULL)
	replyP = reply;
    else
	replyP++;
    printf ("Name server 2:       %s\n", replyP);

    do_query (handle, community, timeout, retries, PKT_GETRQS,
	      "imsConfigTftpServer1", reply, sizeof (reply));
    if ((replyP = strstr (reply, "=")) == NULL)
	replyP = reply;
    else
	replyP++;
    printf ("TFTP server 1:       %s\n", replyP);

    do_query (handle, community, timeout, retries, PKT_GETRQS,
	      "imsConfigTftpServer2", reply, sizeof (reply));
    if ((replyP = strstr (reply, "=")) == NULL)
	replyP = reply;
    else
	replyP++;
    printf ("TFTP server 2:       %s\n", replyP);

    do_query (handle, community, timeout, retries, PKT_GETRQS,
	      "imsConfigLinkopsPort", reply, sizeof (reply));
    if ((replyP = strstr (reply, "=")) == NULL)
	replyP = reply;
    else
	replyP++;
    linkops_port = atoi (replyP);
    if (linkops_port > 0)
    {
	printf ("Linkops port:        %d\n", linkops_port);
    }

    do_query (handle, community, timeout, retries, PKT_GETRQS,
	      "imsConfigLinkSpeed", reply, sizeof (reply));
    if ((replyP = strstr (reply, "=")) == NULL)
	replyP = reply;
    else
	replyP++;
    speed = atoi (replyP);
    if (speed < 1 || speed > (sizeof (speeds) / sizeof (char *)))
    {
	speed = 1;
    }

    printf ("Link speed:          %s\n", speeds[speed - 1]);

    do_query (handle, community, timeout, retries, PKT_GETRQS,
	      "imsConfigTftpBootfile", reply, sizeof (reply));
    if ((replyP = strstr (reply, "=")) == NULL)
	replyP = reply;
    else
	replyP++;
    if (strlen (replyP) > 0 && strcmp (replyP, "\"\""))
	printf ("Boot file:           %s\n", replyP);
    else
	printf ("Network Booting disabled\n");

    do_query (handle, community, timeout, retries, PKT_GETRQS,
	      "imsConfigTftpConfig", reply, sizeof (reply));
    if ((replyP = strstr (reply, "=")) == NULL)
	replyP = reply;
    else
	replyP++;
    tftp_config = atoi (replyP);

    do_query (handle, community, timeout, retries, PKT_GETRQS,
	      "imsConfigUpDateNVRAM", reply, sizeof (reply));
    if ((replyP = strstr (reply, "=")) == NULL)
	replyP = reply;
    else
	replyP++;
    update_nvram = atoi (replyP);

    do_query (handle, community, timeout, retries, PKT_GETRQS,
	      "imsConfigLookupNames", reply, sizeof (reply));
    if ((replyP = strstr (reply, "=")) == NULL)
	replyP = reply;
    else
	replyP++;
    lookup_names = atoi (replyP);

    printf ("Network Configuration ");
    if (tftp_config == 2 && update_nvram == 2)
	printf ("enabled and saved");
    else if (tftp_config == 2)
	printf ("enabled");
    else
	printf ("disabled");
    printf (", DNS lookup ");
    if (lookup_names == 2)
	printf ("enabled\n");
    else
	printf ("disabled\n");
}

void
help_scr (const char *const argv0, const int interactive)
{
    printf ("\nHelp information:\n\n    Key     Action\n");

    printf ("     C      Display ICMP statistics\n");

#ifdef NOTYET
    printf ("     D      Display subsystem names\n");
    printf ("     E      Edit system configuration\n");
#endif				/* NOT YET */

    printf ("     H      Display subsystem history statistics\n");
    printf ("     I      Display IP statistics\n");

#ifdef NOTYET
    printf ("     L      Display continuous receive and transmit statistics\n");
#endif				/* NOT YET */

    printf ("     M      Display memory statistics\n");
    printf ("     N      Display SNMP statistics\n");
    printf ("     P      Display TCP statistics - connection & misc\n");

    if (interactive)
	printf ("     Q      Quit %s\n", argv0);

#ifdef NOTYET
    printf ("     R      Redisplay last 10 messages\n");
#endif				/* NOT YET */

    printf ("     S      Display subsystem information\n");
    printf ("     T      Display TCP statistics - packets sent & receive\n");
    printf ("     U      Display UDP statistics\n");
    printf ("     V      Display version & configuration information\n");
    printf ("     ?      Display this help message\n");
}

char           *SNMP_Errors[] =
{
    "No Error",			/* PKT_NOERROR */
    "To many SNMP varaibles in one request",	/* PKT_TOOBIG */
    "No such MIB variable",	/* PKT_NOSUCHNAME */
    "Illegal value",		/* PKT_BADVALUE */
    "Attempt to write READONLY MIB variable",	/* PKT_READONLY */
    "Unknown SNMP error"	/* PKT_GENERR */
};

void
do_query (mgt_sck * handle,
	  const char *const community,
	  const unsigned long timeout,
	  const int retries,
	  const int type,
	  char *query,
	  char *reply,
	  const int replylen)
{
    pkt_msg         message;
    int             table[256], tableLength, r;

    if (GetList (query, message.MsgPdu.PduRor.RorLst, PKT_SZERORLST,
		 (int *) &message.MsgPdu.PduRor.RorLstLen) < 0)
    {
	fprintf (stderr, "Couldn't parse <%s>\n", query);
	exit (1);
    }

    tableLength = message.MsgPdu.PduRor.RorLst[0].ObjOjiLen;
    memcpy (table, message.MsgPdu.PduRor.RorLst[0].ObjOji,
	    tableLength * sizeof (int));

    strcpy (message.MsgCom, community);
    message.MsgComLen = strlen (community);

    message.MsgPdu.PduRor.RorTyp = type;
    message.MsgPdu.PduRor.RorRid = 0;
    message.MsgPdu.PduRor.RorErrSts = PKT_NOERROR;
    message.MsgPdu.PduRor.RorErrInd = 0;

    if (MgtRqs (handle, &message, timeout, retries) < 0)
    {
	fprintf (stderr, "SNMP request failed, to many retries\n");
	exit (1);
    }

    if (message.MsgPdu.PduRor.RorErrSts != PKT_NOERROR)
    {
	if ((r = PutIdentifier (message.MsgPdu.PduRor.RorLst +
				message.MsgPdu.PduRor.RorErrInd - 1,
				reply, replylen)) < 0)
	{
	    reply[0] = '\0';
	}
	else
	{
	    reply[r] = '\0';
	}
	fprintf (stderr, "SNMP error: <%s>, variable %d, %s\n",
		 SNMP_Errors[message.MsgPdu.PduRor.RorErrSts],
		 message.MsgPdu.PduRor.RorErrInd, reply);
	exit (1);
    }

    if (PutList (message.MsgPdu.PduRor.RorLst, message.MsgPdu.PduRor.RorLstLen,
		 reply, replylen) < 0)
    {
	fprintf (stderr, "Couldn't handle result list\n");
	exit (1);
    }
}

void
ImsFullMib (void)
{
    int             index;
    char            line[256];

    index = 0;
    while (mibdefs[index] != NULL)
    {
	strcpy (line, mibdefs[index]);
	(void) MibScanLine (line);

	index++;
    }
}
